# -*- coding: utf-8 -*-

import logging
from time import sleep
import requests
from requests.exceptions import JSONDecodeError


from src.exceptions import (
    HTTPRequestError,
    UnexpectedResponseError,
    MaxAttemptError,
)
from src.globals import MAX_ATTEMPT, ATTEMPT_RATE


def multiple_attempt(call_attempt=None):
    def wrap(*args, **kwargs):
        count = 0
        while True:
            try:
                return call_attempt(*args, **kwargs)
            except Exception as exc:
                if count < MAX_ATTEMPT:
                    logging.info(f"Call failed {count} times, will retry...")
                    sleep(ATTEMPT_RATE)
                    count += 1
                else:
                    raise MaxAttemptError(f"{call_attempt} Call Error") from exc

    return wrap


@multiple_attempt
def http_request(func):
    def wrap(*args, **kwargs):
        try:
            url, full = func(*args, **kwargs)
            res = requests.get(url, timeout=10)
            if res.status_code == 200:
                try:
                    if full:
                        return res.json()
                    else:
                        return res.json()["data"]
                except JSONDecodeError:
                    return res.content
            else:
                raise HTTPRequestError(res.status_code, res.reason, res.text)
        except Exception as exc:
            raise UnexpectedResponseError from exc

    return wrap



# -*- coding: utf-8 -*-

from os import path
import json
from src.globals.network import Network


def _get_network_path(network: Network):
    # Get the path to the network's ABI folder
    script_dir = path.dirname(path.dirname(__file__))
    return path.join(path.join(script_dir, "abis"), network.name)


def _open_abi_file(network: Network, name: str, folder: str):
    # Load the specified ABI file for the given network and contract/module name
    abi_suffix = ".json"
    abi_path = path.join(_get_network_path(network), folder, name + abi_suffix)
    with open(abi_path, "r", encoding="utf-8") as file:
        a = file.read()
    return json.loads(a)


def get_contract_abi(network: Network, kind: str, name: str):
    # Get the ABI for the specified contract name on the given network.
    # type should be middleware, package or token.
    abi = _open_abi_file(network, name, folder=kind)
    return (abi["address"], abi["abi"])



# -*- coding: utf-8 -*-

from typing import List, Any
from web3 import Web3


def keccak256(x: bytes) -> bytes:
    # Calculate Keccak-256 hash of the input bytes
    return Web3.keccak(x)


def solidity_keccak256(types: List[str], params: List[Any]):
    # Calculate Keccak-256 hash of the given solidity types and parameters
    return Web3.solidity_keccak(abi_types=types, values=params)



# -*- coding: utf-8 -*-

import os

# Solidity
ZERO_BYTES = b""
ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"
ZERO_BYTES32 = b"\x00" * 32

# 100% = 1e10 on Geode's contracts
PERCENTAGE_DENOMINATOR = 10_000_000_000

# 1 ether = 1e18 wei
ETHER_DENOMINATOR: int = 1_000_000_000_000_000_000

# beacon has 9 decimal points, while eth has 18
ETH2GWEI = 1_000_000_000

# 1 execution ether = 1 beacon ether * 1e9
BEACON_DENOMINATOR: int = 1_000_000_000

# Validator Constants
DEPOSIT_CLI_VERSION = "2.4.0"
WORD_LIST_PATH = os.path.join("geode", "utils", "bls")

# Execution-spec constants taken from:
# https://github.com/ethereum/consensus-specs/blob/dev/specs/phase0/beacon-chain.md
DOMAIN_DEPOSIT = bytes.fromhex("03000000")
BLS_WITHDRAWAL_PREFIX = bytes.fromhex("00")
ETH1_ADDRESS_WITHDRAWAL_PREFIX = bytes.fromhex("01")

MIN_DEPOSIT_AMOUNT = 2**0 * ETHER_DENOMINATOR
MAX_DEPOSIT_AMOUNT = 2**5 * ETHER_DENOMINATOR

# Network
REFRESH_RATE = 60  # in seconds
MAX_ATTEMPT = 60
ATTEMPT_RATE = 0.1  # in seconds



# -*- coding: utf-8 -*-

import logging
from src.globals import Network
from src.utils import http_request


class Beacon:
    """
    https://ethereum.github.io/beacon-APIs/#
    Most of the GET end points that is specified within
    the Ethereum's Specification v2.3.0 is accesible.
    Some of the v2.4.0 is accesible too.
    """

    def __init__(self, network: Network, cons_api: str):
        """
        Initializes a new instance of the class.

        Args:
            network: A `Network` object representing the blockchain network to connect to.
            cons_api: API that is compatible with beacon-APIs specification.
            **kwargs: Any additional keyword arguments to pass to the initializer.
        """
        # Get the API base URL for the specified `network`
        self.api_base: str = cons_api
        self.network: Network = network

        # Log a message indicating that the class instance is connected to the API base
        logging.info(f"Connected to beconchain")

    @http_request
    def beacon_genesis(self):
        # Retrieve details of the chain's genesis which can be used to identify chain.
        url = self.api_base + f"/eth/v1/beacon/genesis"
        return (url, False)

    @http_request
    def beacon_states_root(self, state_id: str):
        # Calculates HashTreeRoot for state with given 'stateId'.
        # If stateId is root, same value will be returned.
        url = self.api_base + f"/eth/v1/beacon/states/{state_id}/root"
        return (url, False)

    @http_request
    def beacon_states_fork(self, state_id: str):
        # Returns Fork object for state with given 'stateId'.
        url = self.api_base + f"/eth/v1/beacon/states/{state_id}/fork"
        return (url, False)

    @http_request
    def beacon_states_finality_checkpoints(self, state_id: str):
        # Returns finality checkpoints for state with given 'stateId'.
        # In case finality is not yet achieved,
        # checkpoint should return epoch 0 and ZERO_HASH as root.
        url = (
            self.api_base
            + f"/eth/v1/beacon/states/{state_id}/finality_checkpoints"
        )
        return (url, False)

    @http_request
    def beacon_states_validators(self, state_id: str):
        # Returns filterable list of validators with their balance, status and index.
        # Information will be returned for all indices or public key that match known validators.
        # If an index or public key does not match any known validator,
        # no information will be returned but this will not cause an error.
        # There are no guarantees for the returned data in terms of ordering;
        # both the index and public key are returned for each validator,
        # and can be used to confirm for which inputs a response has been returned.
        url = self.api_base + f"/eth/v1/beacon/states/{state_id}/validators"
        return (url, False)

    @http_request
    def beacon_states_validators_id(self, state_id: str, validator_id: str):
        # Returns validator specified by state and id or public key along with status and balance.
        url = (
            self.api_base
            + f"/eth/v1/beacon/states/{state_id}/validators/{validator_id}"
        )
        return (url, False)

    @http_request
    def beacon_states_validator_balances(self, state_id: str):
        # Returns filterable list of validators balances.
        url = (
            self.api_base
            + f"/eth/v1/beacon/states/{state_id}/validator_balances"
        )
        return (url, False)

    @http_request
    def beacon_states_committees(self, state_id: str):
        # Retrieves the committees for the given state.
        url = self.api_base + f"/eth/v1/beacon/states/{state_id}/committees"
        return (url, False)

    @http_request
    def beacon_states_sync_committees(self, state_id: str):
        # Retrieves the current sync committee for the given state.
        # Also returns the subcommittee assignments.
        url = (
            self.api_base + f"/eth/v1/beacon/states/{state_id}/sync_committees"
        )
        return (url, False)

    @http_request
    def beacon_states_randao(self, state_id: str):
        # Fetch the RANDAO mix for the requested epoch from the state identified by state_id.
        # If an epoch is not specified then the RANDAO mix for the state's
        # current epoch will be returned.
        # By adjusting the state_id parameter you can query for any
        # historic value of the RANDAO mix.
        # Ordinarily states from the same epoch will mutate the RANDAO mix
        # for that epoch as blocks are applied
        url = self.api_base + f"/eth/v1/beacon/states/{state_id}/randao"
        return (url, False)

    @http_request
    def beacon_headers(self):
        # Retrieves block header for given block id.
        url = self.api_base + f"/eth/v1/beacon/headers"
        return (url, False)

    @http_request
    def beacon_headers_id(self, block_id: str):
        # Retrieves block header for given block id.
        url = self.api_base + f"/eth/v1/beacon/headers/{block_id}"
        return (url, False)

    @http_request
    def beacon_blocks(self, block_id: str):
        # Retrieves block details for given block id.
        url = self.api_base + f"/eth/v2/beacon/blocks/{block_id}"
        return (url, False)

    @http_request
    def beacon_blinded_blocks_(self, block_id: str):
        # Retrieves blinded block for given block ID.
        url = self.api_base + f"/eth/v1/beacon/blinded_blocks/{block_id}"
        return (url, False)

    @http_request
    def beacon_blocks_id(self, block_id: str):
        # Retrieves hashTreeRoot of BeaconBlock/BeaconBlockHeader
        url = self.api_base + f"/eth/v1/beacon/blocks/{block_id}"
        return (url, False)

    @http_request
    def beacon_blocks_root(self, block_id: str):
        # Retrieves hashTreeRoot of BeaconBlock/BeaconBlockHeader
        url = self.api_base + f"/eth/v1/beacon/blocks/{block_id}/root"
        return (url, False)

    @http_request
    def beacon_blocks_attestations(self, block_id: str):
        # Retrieves attestation included in requested block.
        url = self.api_base + f"/eth/v1/beacon/blocks/{block_id}/attestations"
        return (url, False)

    @http_request
    def beacon_blob_sidecars(self, block_id: str):
        # Retrieves blob sidecars for a given block id.
        # When indices is specified, only the blob sidecars with the specified indices are returned.
        # There are no guarantees for the returned blob sidecars in terms of ordering.
        url = self.api_base + f"/eth/v1/beacon/blob_sidecars/{block_id}"
        return (url, False)

    @http_request
    def beacon_pool_attestations(self):
        # Retrieves attestations known by the node but not necessarily incorporated into any block
        url = self.api_base + f"/eth/v1/beacon/pool/attestations"
        return (url, False)

    @http_request
    def beacon_pool_attester_slashings(self):
        # Retrieves attester slashings known by the node
        # but not necessarily incorporated into any block
        url = self.api_base + f"/eth/v1/beacon/pool/attester_slashings"
        return (url, False)

    @http_request
    def beacon_pool_proposer_slashings(self):
        # Retrieves proposer slashings known by the node
        # but not necessarily incorporated into any block
        url = self.api_base + f"/eth/v1/beacon/pool/proposer_slashings"
        return (url, False)

    @http_request
    def beacon_pool_voluntary_exits(self):
        # Retrieves voluntary exits known by the node but
        # not necessarily incorporated into any block
        url = self.api_base + f"/eth/v1/beacon/pool/voluntary_exits"
        return (url, False)

    @http_request
    def beacon_pool_bls_to_execution_changes(self):
        # Retrieves BLS to execution changes known by the
        # node but not necessarily incorporated into any block
        url = self.api_base + f"/eth/v1/beacon/pool/bls_to_execution_changes"
        return (url, False)

    @http_request
    def beacon_deposit_snapshot(self):
        # Retrieve EIP-4881 Deposit Tree Snapshot.
        url = self.api_base + f"/eth/v1/beacon/deposit_snapshot"
        return (url, False)

    @http_request
    def builder_states_expected_withdrawals(self, state_id: str):
        # Get the withdrawals computed from the specified state,
        # that will be included in the block that gets built on the specified state.
        url = (
            self.api_base
            + f"/eth/v1/builder/states/{state_id}/expected_withdrawals"
        )
        return (url, False)

    @http_request
    def config_fork_schedule(self):
        # Retrieve all forks, past present and future,
        # of which this node is aware.
        url = self.api_base + f"/eth/v1/config/fork_schedule"
        return (url, False)

    @http_request
    def config_spec(self):
        # Retrieve specification configuration used on this node
        url = self.api_base + f"/eth/v1/config/spec"
        return (url, False)

    @http_request
    def config_deposit_contract(self):
        # Retrieve Eth1 deposit contract address and chain ID.
        url = self.api_base + f"/eth/v1/config/deposit_contract"
        return (url, False)

    @http_request
    def debug_beacon_states(self, state_id: str):
        # Returns full BeaconState object for given stateId.
        url = self.api_base + f"/eth/v2/debug/beacon/states/{state_id}"
        return (url, True)

    @http_request
    def debug_beacon_heads(self):
        # Retrieves all possible chain heads (leaves of fork choice tree).
        url = self.api_base + f"/eth/v2/debug/beacon/heads"
        return (url, True)

    @http_request
    def debug_fork_choice(self):
        # Retrieves all current fork choice context.
        url = self.api_base + f"/eth/v1/debug/fork_choice"
        return (url, True)

    # @http_request
    # def events(self, topic: str):
    # This functionality is currently not available, and will be implemented later.
    # Keeping it as a TODO.
    # Provides endpoint to subscribe to beacon node Server-Sent-Events stream.
    # Consumers should use eventsource implementation to listen on those events.
    # Servers may send SSE comments beginning with : for any purpose,
    # including to keep the event stream connection alive in the presence of proxy servers.
    # https://html.spec.whatwg.org/multipage/server-sent-events.html the-eventsource-interface
    # Event types to subscribe to:
    # head, block, attestation, voluntary_exit, finalized_checkpoint, chain_reorg, contribution_and_proof
    #     url = self.api_base + f"/eth/v1/events?topics={topic}"
    #     print(url)
    #     return (url, True)

    @http_request
    def node_peers(self):
        # Retrieves data about the node's network peers.# By default this returns all peers.
        # Multiple query params are combined using AND conditions
        url = self.api_base + f"/eth/v1/node/peers"
        return (url, False)

    @http_request
    def node_peers_id(self, peer_id: str):
        # Retrieves data about the given peer
        url = self.api_base + f"/eth/v1/node/peers/{peer_id}"
        return (url, False)

    @http_request
    def node_peer_count(self):
        # Retrieves number of known peers.
        url = self.api_base + f"/eth/v1/node/peer_count"
        return (url, False)

    @http_request
    def node_version(self):
        # Requests that the beacon node identify information about its implementation
        # in a format similar to a HTTP User-Agent field.
        url = self.api_base + f"/eth/v1/node/version"
        return (url, False)

    @http_request
    def node_syncing(self):
        # Requests the beacon node to describe if it's currently syncing or not,
        # and if it is, what block it is up to.
        url = self.api_base + f"/eth/v1/node/syncing"
        return (url, False)

    @http_request
    def node_health(self):
        # Returns node health status in http status codes.
        # Useful for load balancers.
        url = self.api_base + f"/eth/v1/node/health"
        return (url, False)



# -*- coding: utf-8 -*-

from enum import IntEnum
import typing as t

from .network import Network


class DEPOSIT_SIZE(IntEnum):
    """
    proposeStake: takes 1 ether from the operator internal wallet
    beaconStake: takes 31 ether from pool & reimburses the operator's 1 ether
    """

    PROPOSAL = 1_000_000_000
    STAKE = 31_000_000_000


DEPOSIT_NETWORK_NAME: t.Dict[Network, bytes] = {
    Network.ethereum: "mainnet",
    Network.holesky: "holesky",
    Network.binance: None,
    Network.gnosis: "gnosis",
    Network.avalanche: None,
}

GENESIS_FORK_VERSION: t.Dict[Network, bytes] = {
    Network.ethereum: bytes.fromhex("00000000"),
    Network.holesky: bytes.fromhex("01017000"),
    Network.binance: None,
    Network.gnosis: bytes.fromhex("00000064"),
    Network.avalanche: None,
}



# -*- coding: utf-8 -*-

import logging
from web3 import Web3
from web3.contract import Contract

from src.globals import Network
from src.utils import get_contract_abi, to_bytes32, to_string
from src.exceptions import UnknownChainError

from .pool import Pool
from .operator import Operator
from .beacon import Beacon
from .validator import Validator


class Portal:
    """
    Attributes:
        w3: A Web3 instance used for interacting with the Ethereum network.
        network: A Network instance representing the Ethereum network to use.
        portal: An Ethereum address of the portal contract.
    """

    w3: Web3  # Web3 instance used for interacting with the Ethereum network
    network: Network  # Network to which the Portal is connected
    contract: Contract  # Contract instance used for interacting with the Portal contract
    beacon: Beacon  # Beacon instance used for generating deposit keys and signatures
    version: int  # Version of the Portal contract

    def __init__(self, w3: Web3, beacon: Beacon):
        """
        Initializes a new Portal instance.

        Args:
            w3 (Web3): Web3 instance used for interacting with the Ethereum network.
            beacon (Beacon): Beacon instance used for generating deposit keys and signatures.
            kwargs (dict): Optional keyword arguments.
        """
        self._set_web3(w3)  # sets the Web3 instance for the Portal
        if (
            self.network is Network.ethereum
            or self.network is Network.holesky
            or self.network is Network.gnosis
        ):
            self._set_beacon(beacon)
        else:
            raise UnknownChainError

        address, abi = get_contract_abi(
            network=self.network, kind="package", name="Portal"
        )
        self.contract: Contract = w3.eth.contract(
            address=Web3.to_checksum_address(address), abi=abi
        )

        self.version: int = self.functions.getContractVersion().call()
        version_name = self.functions.readBytes(
            self.version, to_bytes32("NAME")
        ).call()
        logging.info(
            f"Portal:{self.network.name} head is on '{to_string(version_name)}'"
        )

    def _set_web3(self, w3: Web3):
        """
        Set the Web3 instance and network based on the chain ID.
        """

        self.w3: Web3 = w3
        self.network: Network = Network(self.w3.eth.chain_id)

    def _set_beacon(self, beacon: Beacon):
        """
        Set the Beacon instance
        """
        self.beacon: Beacon = beacon

    def __getattr__(self, attr):
        return getattr(self.contract, attr)

    def pool(self, id_: int):
        """
        Get the Pool object by id
        """
        return Pool(
            beacon=self.beacon,
            w3=self.w3,
            network=self.network,
            portal=self.contract,
            id_=id_,
        )

    def operator(self, id_: int):
        """
        Get the Opeartor object by id
        """
        return Operator(
            w3=self.w3, network=self.network, portal=self.contract, id_=id_
        )

    def validator(self, pubkey: bytes):
        """
        Get the Opeartor object by id
        """
        return Validator(
            w3=self.w3,
            network=self.network,
            portal=self.contract,
            beacon=self.beacon,
            pk=pubkey,
        )



# -*- coding: utf-8 -*-


class PythonVersionError(Exception):
    "Python version is not supported."


class UnknownChainError(Exception):
    "Provided RPC points to an unknown chainId"


class UnexpectedResponseError(Exception):
    "Https call resulted with an error code (non 2xx)"


class HTTPRequestError(Exception):
    "Response code is not 200"


class MaxAttemptError(Exception):
    "Multiple attempt made but all failed"


class UnknownApiError(Exception):
    "Provided API endpoint is not recognized"



# -*- coding: utf-8 -*-

from src.globals import ID_TYPE

from .id import Id


class Operator(Id):
    def __init__(self, *args, **kwargs):
        """
        Initializes an Operator object.

        Parameters:
            *args: variable length argument list.
            **kwargs: arbitrary keyword arguments.
        """
        # The type of the Operator is 4
        super().__init__(*args, **kwargs, type_=ID_TYPE(4))

    @property
    def initiated(self):
        return self._read_uint("initiated")

    @property
    def maintainer(self):
        return self._read_address("maintainer")

    @property
    def feeSwitch(self):
        return self._read_uint("feeSwitch")

    @property
    def priorFee(self):
        return self._read_uint("priorFee")

    @property
    def fee(self):
        return self._read_uint("fee")

    @property
    def periodSwitch(self):
        return self._read_uint("periodSwitch")

    @property
    def priorPeriod(self):
        return self._read_uint("priorPeriod")

    @property
    def validatorPeriod(self):
        return self._read_uint("validatorPeriod")

    @property
    def wallet(self):
        return self._read_uint("wallet")

    @property
    def release(self):
        return self._read_uint("release")



# -*- coding: utf-8 -*-

import codecs


def to_bytes(key: str):
    encoded = codecs.encode(key.encode("utf-8"), "hex_codec")
    return str("0x" + encoded.decode("utf-8"))


def to_bytes32(key: str):
    encoded = codecs.encode(key.encode("utf-8"), "hex_codec")
    padded = encoded + b"0" * (64 - len(encoded))
    return str("0x" + padded.decode("utf-8"))


def to_string(key: bytes):
    return key.decode()



# -*- coding: utf-8 -*-

import logging
import typing as t
from datetime import datetime, timedelta
from web3 import Web3
from web3.contract import Contract

from src.globals import Network, VALIDATOR_STATE, REFRESH_RATE
from src.utils import multiple_attempt

from .beacon import Beacon


# This is a static method and a decorator that wraps around another method
# It is used to update the Portal state for the Validator
@multiple_attempt
def update_portal(func):
    # This is the wrapper function that is returned by the decorator.
    # It updates the Portal state if REFRESH_RATE has elapsed
    def wrap(self):
        if datetime.now() > self.last_portal_update + timedelta(
            seconds=REFRESH_RATE
        ):
            # Call the `getValidator()` function on the Portal contract with Validator's pubkey
            # and retrieve its state
            self.portal_state = self.portal.functions.getValidator(
                self.pubkey
            ).call()
            self.last_portal_update = datetime.now()
        # Call the function that was decorated and return its result
        return func(self)

    # Return the wrapper function
    return wrap


# This is a static method and a decorator that wraps around another method
# It is used to update the Beacon chain state for the Validator
@multiple_attempt
def update_beacon(func):
    # This is a decorator function that wraps around the given function.
    # It checks whether the last update to the beacon state is older than REFRESH_RATE seconds.
    def wrap(self):
        if datetime.now() > self.last_beacon_update + timedelta(
            seconds=REFRESH_RATE
        ):
            # If it is, it updates the beacon state by calling get_validator()
            # function of the beacon contract.
            self.beacon_state = self.beacon.beacon_states_validators_id(
                state_id="head", validator_id=self.pubkey
            )
            # It then updates the last_beacon_update timestamp to the current time.
            self.last_beacon_update = datetime.now()
        return func(self)

    # Finally, it returns the wrapped function.
    return wrap


class Validator:
    # Declare instance variables for Validator class
    w3: Web3  # An instance of the Web3 class for interacting with Ethereum blockchain
    network: Network  # The network being validated
    portal: Contract  # The portal contract being used
    beacon: Beacon  # The beacon chain being used

    def __init__(
        self, w3: Web3, network: Network, portal: Contract, beacon: Beacon, pk
    ):
        """
        Initialize the Validator object.

        pk: Public key of the validator, bytes (no 0x prefix) or str (0x prefix)
        """
        self.w3 = w3
        self.network = network
        self.portal = portal
        self.beacon = beacon
        if isinstance(pk, bytes):
            self.pubkey: str = "0x" + pk.hex()
        elif isinstance(pk, str):
            self.pubkey: str = pk

        self.portal_state = []  # A list containing the current portal state
        self.beacon_state: t.Dict = (
            {}
        )  # A dictionary containing the current beacon state
        # update timestamps set on init to be able to catch init state without
        self.last_portal_update: datetime = datetime.now() - timedelta(
            seconds=REFRESH_RATE + 1
        )  # The time of the last portal state update
        self.last_beacon_update: datetime = datetime.now() - timedelta(
            seconds=REFRESH_RATE + 1
        )  # The time of the last beacon state update

        logging.info(f"Connected to the validator: {self.pubkey}")

    def __str__(self):
        return f"Validator Object: {self.pubkey}"

    # Portal Props
    @property
    @update_portal
    def portal_index(self):
        return self.portal_state[1]

    @property
    @update_portal
    def state(self):
        return VALIDATOR_STATE(self.portal_state[0])

    @property
    @update_portal
    def createdAt(self):
        return self.portal_state[2]

    @property
    @update_portal
    def period(self):
        return self.portal_state[3]

    @property
    @update_portal
    def poolId(self):
        return self.portal_state[4]

    @property
    @update_portal
    def operatorId(self):
        return self.portal_state[5]

    @property
    @update_portal
    def poolFee(self):
        return self.portal_state[6]

    @property
    @update_portal
    def operatorFee(self):
        return self.portal_state[7]

    @property
    @update_portal
    def infrastructureFee(self):
        return self.portal_state[8]

    @property
    @update_portal
    def signature31(self):
        return self.portal_state[9]

    @property
    @update_portal
    def __portal__(self):
        keys = [
            "state",
            "index",
            "createdAt",
            "period",
            "poolId",
            "operatorId",
            "poolFee",
            "operatorFee",
            "governanceFee",
            "signature31",
        ]

        return {keys[i]: value for i, value in enumerate(self.portal_state)}

    # Beacon Props

    @property
    @update_beacon
    def beacon_index(self):
        return self.beacon_state["index"]

    @property
    @update_beacon
    def balance(self):
        return self.beacon_state["balance"]

    @property
    @update_beacon
    def beacon_status(self):
        return self.beacon_state["status"]

    @property
    @update_beacon
    def withdrawal_credentials(self):
        return self.beacon_state["validator"]["withdrawal_credentials"]

    @property
    @update_beacon
    def effective_balance(self):
        return self.beacon_state["validator"]["effective_balance"]

    @property
    @update_beacon
    def slashed(self):
        return self.beacon_state["validator"]["slashed"]

    @property
    @update_beacon
    def activation_eligibility_epoch(self):
        return self.beacon_state["validator"]["activation_eligibility_epoch"]

    @property
    @update_beacon
    def exit_epoch(self):
        return self.beacon_state["validator"]["exit_epoch"]

    @property
    @update_beacon
    def withdrawable_epoch(self):
        return self.beacon_state["validator"]["withdrawable_epoch"]

    @property
    @update_beacon
    def __beacon__(self):
        return self.beacon_state



# -*- coding: utf-8 -*-

from .id import ID_TYPE
from .network import Network
from .token import TOKEN_NAMES
from .validator import VALIDATOR_STATE
from .beacon import DEPOSIT_SIZE, DEPOSIT_NETWORK_NAME, GENESIS_FORK_VERSION
from .constants import *



# -*- coding: utf-8 -*-

import typing as t

from src.globals import ID_TYPE, DEPOSIT_SIZE
from src.utils import get_key


from .id import Id
from .validator import Validator
from .beacon import Beacon


class Pool(Id):
    def __init__(self, beacon: Beacon, *args, **kwargs):
        """
        Initializes a Pool object.

        Parameters:
            beacon: An initialized Beacon object.
            *args: variable length argument list.
            **kwargs: arbitrary keyword arguments.

        Returns:
        None
        """
        # The type of the Pool is 5.
        super().__init__(*args, **kwargs, type_=ID_TYPE(5))
        # The Pool object is associated with a Beacon object.
        self.beacon: Beacon = beacon

    @property
    def initiated(self):
        return self._read_uint("initiated")

    @property
    def maintainer(self):
        return self._read_address("maintainer")

    @property
    def yieldReceiver(self):
        return self._read_address("yieldReceiver")

    @property
    def surplus(self):
        return self._read_uint("surplus")

    @property
    def secured(self):
        return self._read_uint("secured")

    @property
    def middlewares_list(self):
        lenmiddlewares = self.middlewares_len
        ints: t.List = []
        for i in range(lenmiddlewares):
            ints.append(self._read_address_array("middlewares", i))
        return ints

    @property
    def middlewares_len(self):
        return self._read_uint("middlewares")

    def middlewares(self, index):
        return self._read_address_array("middlewares", index)

    @property
    def private(self):
        return self._read_uint("private") == 1

    @property
    def whitelist(self):
        return self._read_address("whitelist")

    @property
    def withdrawalCredential(self):
        return self._read_bytes("withdrawalCredential", is_hex=True)

    @property
    def withdrawalContract(self):
        return self._read_address("withdrawalContract")

    @property
    def liquidityPool(self):
        return self._read_address("liquidityPool")

    @property
    def feeSwitch(self):
        return self._read_uint("feeSwitch")

    @property
    def priorFee(self):
        return self._read_uint("priorFee")

    @property
    def fee(self):
        return self._read_uint("fee")

    @property
    def wallet(self):
        return self._read_uint("wallet")

    @property
    def validatorsList(self):
        lenvalidators = self.validatorsLen
        vals: t.List = []
        for i in range(lenvalidators):
            vals.append(self._read_bytes_array("validators", index=i))
        return vals

    @property
    def validatorsLen(self):
        return self._read_uint("validators")

    @property
    def fallbackOperator(self):
        return self._read_uint("fallbackOperator")

    @property
    def fallbackThreshold(self):
        return self._read_uint("fallbackThreshold")

    def validators(self, index):
        return Validator(
            network=self.network,
            beacon=self.beacon,
            w3=self.w3,
            portal=self.portal,
            pk=self._read_bytes_array("validators", index=index),
        )

    def allowance(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="allowance"))

    def proposedValidators(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="proposedValidators"))

    def activeValidators(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="activeValidators"))

    def alienValidators(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="alienValidators"))



# -*- coding: utf-8 -*-

import typing as t
from src.globals.network import Network


TOKEN_NAMES: t.Dict[Network, str] = {
    Network.ethereum: "gETH",
    Network.holesky: "gETH",
    Network.binance: "gBNB",
    Network.gnosis: "gGNO",
    Network.avalanche: "gAVAX",
}



# -*- coding: utf-8 -*-

from web3 import Web3
from eth_abi import encode

from src.globals import ID_TYPE

from .solidity import to_bytes32
from .crypto import solidity_keccak256, keccak256


def generate_id(name: str, type_: ID_TYPE):
    """
    Generates an ID using keccak256 hash function.

    Parameters:
    name (str): The name used to generate the ID.
    type (ID_TYPE): The type of ID to generate.

    Returns:
    int: The generated ID.
    """
    return Web3.to_int(keccak256(encode(["string", "uint256"], [name, type_])))


def get_key(id_: int, key: str):
    """
    Generates a key using keccak256 hash function.

    Parameters:
    id (int): The ID to use in generating the key.
    key (str): The key to generate.

    Returns:
    bytes: The generated key.
    """

    return solidity_keccak256(["uint256", "bytes32"], [id_, to_bytes32(key)])



# -*- coding: utf-8 -*-

from web3 import Web3, HTTPProvider, WebsocketProvider
from src.exceptions import UnknownChainError, UnknownApiError
from src.globals import Network
from src.utils import check_python_version
from src.classes import Portal, Token, Beacon


class Geode:
    def __init__(self, exec_api: str = "", cons_api: str = ""):
        # Check if the current version of Python is supported
        check_python_version()

        # Set the Web3 instance with the given execution API
        self._set_web3(exec_api)

        # If the network is Ethereum, Holesky or Gnosis, set the Beacon instance with the given key
        if (
            self.network is Network.ethereum
            or self.network is Network.holesky
            or self.network is Network.gnosis
        ):
            self._set_beacon(cons_api)
        else:
            raise UnknownChainError

        # Set the Token instance
        self._set_token()

        # Set the Portal instance
        self._set_portal()

    # Internal method to set the Web3 instance
    def _set_web3(self, exec_api: str):
        if exec_api:
            if exec_api.startswith("https"):
                self.w3: Web3 = Web3(HTTPProvider(exec_api))
            elif exec_api.startswith("wss"):
                self.w3: Web3 = Web3(WebsocketProvider(exec_api))
            else:
                raise UnknownApiError

            self.network: int = Network(self.w3.eth.chain_id)

    # Internal method to set the Beacon instance
    def _set_beacon(self, cons_api: str):
        if cons_api:
            self.beacon: Beacon = Beacon(
                network=self.network, cons_api=cons_api
            )
        else:
            raise UnknownApiError

    # Internal method to set the Portal instance
    def _set_portal(self):
        self.portal: Portal = Portal(self.w3, self.beacon)

    # Internal method to set the Token instance
    def _set_token(self):
        self.token: Token = Token(self.w3, self.network)



# -*- coding: utf-8 -*-

from enum import IntEnum


class VALIDATOR_STATE(IntEnum):
    # invalid
    NONE = 0
    # validator is proposed, 1 ETH is sent from Operator to Deposit Contract.
    PROPOSED = 1
    # proposal was approved, operator used pooled funds, 1 ETH is released back to Operator.
    ACTIVE = 2
    # validator is called to be exited.
    EXIT_REQUESTED = 3
    # validator is fully exited.
    EXITED = 4
    # proposal was malicious(alien). Maybe faulty signatures
    # or probably frontrunning (https://bit.ly/3Tkc6UC)
    ALIENATED = 69



# -*- coding: utf-8 -*-

import sys
from src.exceptions import PythonVersionError


def check_python_version() -> None:
    """Checks that the python version running is sufficient and exits if not."""

    if sys.version_info <= (3, 8) and sys.version_info >= (3, 12):
        raise PythonVersionError(
            f"Python version is not supported.Please consider using 3.8-3.12"
        )



# -*- coding: utf-8 -*-

from .abi import get_contract_abi
from .token import get_token_name
from .solidity import to_bytes, to_bytes32, to_string
from .crypto import solidity_keccak256
from .id import generate_id, get_key
from .wrappers import http_request, multiple_attempt
from .version import check_python_version



# -*- coding: utf-8 -*-

import logging

from web3 import Web3
from web3.contract import Contract

from src.globals import Network
from src.utils import get_contract_abi, get_token_name


class Token:
    network: Network
    contract: Contract

    def __init__(self, w3: Web3, network: Network):
        # Set the network of the token
        self.network: Network = network

        # Get the name, address and ABI of the token contract
        token_name = get_token_name(self.network)
        address, abi = get_contract_abi(
            network=self.network, kind="token", name=token_name
        )

        # Instantiate the contract using Web3 instance, address and ABI
        self.contract: Contract = w3.eth.contract(
            address=Web3.to_checksum_address(address), abi=abi
        )

        # Log the token name
        logging.info(f"Token:{token_name}")

    def __getattr__(self, attr):
        return getattr(self.contract, attr)



# -*- coding: utf-8 -*-

import logging
import eth_typing as et
from web3 import Web3
from web3.contract import Contract

from src.globals import Network, ID_TYPE
from src.utils import to_bytes32, to_string, multiple_attempt
from src.exceptions import UnexpectedResponseError


class Id:
    """
    A class representing an identifier for some object or entity.

    Attributes:
        w3: A Web3 instance used for interacting with the Ethereum network.
        network: A Network instance representing the Ethereum network to use.
        portal: An Ethereum address of the portal contract.
    """

    w3: Web3
    network: Network
    portal: et.ChecksumAddress

    def __init__(
        self,
        w3: Web3,
        network: Network,
        portal: Contract,
        id_: int,
        type_: int,
    ):
        """
        Initializes a new Id instance with the specified parameters.

        Args:
            w3: A Web3 instance used for interacting with the Ethereum network.
            network: A Network instance representing the Ethereum network to use.
            portal: An Ethereum address of the portal contract.
            id: An integer representing the identifier value.
        """
        self.w3 = w3
        self.network = network
        self.portal = portal
        self.ID = id_
        # TYPE: An enumeration value representing the type of the identifier.
        self.TYPE = type_
        logging.info(f"ID TYPE:{ID_TYPE(self.TYPE).name}:{self.ID}")

    @multiple_attempt
    def _read_uint(self, key: str):
        res = self.portal.functions.readUint(self.ID, to_bytes32(key)).call()
        return res

    @multiple_attempt
    def _read_uint_array(self, key: str, index: int):
        res = self.portal.functions.readUintArray(
            self.ID, to_bytes32(key), index
        ).call()
        return res

    @multiple_attempt
    def _read_bytes(self, key: str, is_string: bool = False, is_hex=False):
        res = self.portal.functions.readBytes(self.ID, to_bytes32(key)).call()

        if isinstance(res, str):
            return res
        elif isinstance(res, bytes):
            if is_string:
                return to_string(res)
            elif is_hex:
                return "0x" + res.hex()
            else:
                return res
        else:
            raise UnexpectedResponseError

    @multiple_attempt
    def _read_bytes_array(
        self, key: str, index: int, is_string: bool = False, is_hex=False
    ):
        res = self.portal.functions.readBytesArray(
            self.ID, to_bytes32(key), index
        ).call()

        if isinstance(res, str):
            return res
        elif isinstance(res, bytes):
            if is_string:
                return to_string(res)
            elif is_hex:
                return res.hex()
            else:
                return res
        else:
            raise UnexpectedResponseError

    @multiple_attempt
    def _read_address(self, key: str):
        res: et.Address = self.portal.functions.readAddress(
            self.ID, to_bytes32(key)
        ).call()
        cs_res: et.ChecksumAddress = Web3.to_checksum_address(res)
        return cs_res

    @multiple_attempt
    def _read_address_array(self, key: str, index: int):
        res: et.Address = self.portal.functions.readAddressArray(
            self.ID, to_bytes32(key), index
        ).call()
        cs_res: et.ChecksumAddress = Web3.to_checksum_address(res)
        return cs_res

    @property
    def NAME(self):
        """
        Returns the name of the object or entity.
        """
        return self._read_bytes("NAME", is_string=True)

    @property
    def CONTROLLER(self):
        """
        Returns the Ethereum address of the controller of the object or entity.
        """
        return self._read_address("CONTROLLER")

    def __str__(self):
        """
        Returns a string representation of this Id instance.
        """
        return str(self.ID)



# -*- coding: utf-8 -*-

from enum import IntEnum


class ID_TYPE(IntEnum):
    NONE = 0
    SENATE = 1
    LIMIT_MIN_USER = 3
    OPERATOR = 4
    POOL = 5
    LIMIT_MAX_USER = 9999
    LIMIT_MIN_PACKAGE = 10000
    PACKAGE_PORTAL = 10001
    PACKAGE_WITHDRAWAL_CONTRACT = 10011
    PACKAGE_LIQUIDITY_POOL = 10021
    LIMIT_MAX_PACKAGE = 19999
    LIMIT_MIN_MIDDLEWARE = 20000
    MIDDLEWARE_GETH = 20011
    LIMIT_MAX_MIDDLEWARE = 2999
